# frozen_string_literal: true

require 'hashdiff/util'
require 'hashdiff/compare_hashes'
require 'hashdiff/lcs'
require 'hashdiff/lcs_compare_arrays'
require 'hashdiff/linear_compare_array'
require 'hashdiff/diff'
require 'hashdiff/patch'
require 'hashdiff/version'
